unit UBFTarget;

{$mode delphi}

interface

uses
  Classes, SysUtils, UTarget;

type

  TMachineCommand = (cmNoOperation, cmInc, cmDec, cmPointerInc, cmPointerDec,
    cmOutput, cmInput, cmLoopStart, cmLoopEnd, cmDebug, cmSet, cmMultipy);

  { TMachineOperation }

  TMachineOperation = record
    Command: TMachineCommand;
    Parameter: Integer;
    RelIndex: Integer;
    class function Create(Command: TMachineCommand; Parameter, RelIndex: Integer): TMachineOperation; static;
  end;

  TOptimizations = record
    AddSub: Boolean;
    Merge: Boolean;
    RelativeIndexes: Boolean;
    CopyMultiply: Boolean;
  end;

  { TBFTarget }

  TBFTarget = class(TTarget)
  private
    function CheckClear: Boolean;
    function CheckOccurenceSumParam(C: TMachineCommand): Integer;
    function CheckOccurence(C: TMachineCommand): Integer;
    procedure OptimizeAddSub;
    procedure OptimizeMerge;
    procedure OptimizeZeroInitMemory;
    procedure OptimizeRelativeIndexes;
    procedure OptimizeCopyMultiply;
  protected
    FProgram: array of TMachineOperation;
    FProgramIndex: Integer;
    function GetOperationText(Operation: TMachineOperation): string; virtual;
    procedure LoadProgram; override;
  public
    MemorySize: Integer;
    MemoryMaxUsedAddr: Integer;
    CellSize: Integer;
    Optimizations: TOptimizations;
    constructor Create; override;
    procedure OptimizeSource; override;
    property ProgramIndex: Integer read FProgramIndex;
  end;

const
  BrainFuckCommandText: array[TMachineCommand] of Char = (
    ' ', '+', '-', '>', '<', '.', ',', '[', ']', '@', '=', '*');


implementation

{ TMachineOperation }

class function TMachineOperation.Create(Command: TMachineCommand; Parameter,
  RelIndex: Integer): TMachineOperation;
begin
  Result.Command := Command;
  Result.Parameter := Parameter;
  Result.RelIndex := RelIndex;
end;

function TBFTarget.CheckClear: Boolean;
begin
  Result := (FProgram[FProgramIndex].Command = cmLoopStart) and (Length(FProgram) >= FProgramIndex + 2) and
    (((FProgram[FProgramIndex + 1].Command = cmDec) and (FProgram[FProgramIndex + 1].Parameter = 1)) or
    ((FProgram[FProgramIndex + 1].Command = cmInc) and (FProgram[FProgramIndex + 1].Parameter = -1)))
    and (FProgram[FProgramIndex + 2].Command = cmLoopEnd);
end;

function TBFTarget.CheckOccurence(C: TMachineCommand): Integer;
begin
  Result := 1;
  while ((FProgramIndex + 1) < Length(FProgram)) and (FProgram[FProgramIndex + 1].Command = C) do begin
    Inc(Result);
    Inc(FProgramIndex);
  end;
end;

function TBFTarget.CheckOccurenceSumParam(C: TMachineCommand): Integer;
begin
  Result := FProgram[FProgramIndex].Parameter;
  while ((FProgramIndex + 1) < Length(FProgram)) and (FProgram[FProgramIndex + 1].Command = C) do begin
    Inc(Result, FProgram[FProgramIndex + 1].Parameter);
    Inc(FProgramIndex);
  end;
end;

procedure TBFTarget.OptimizeAddSub;
var
  NewProgram: array of TMachineOperation;
  NewProgramIndex: Integer;
  NewTargetPos: Integer;
  FirstIndex: Integer;
begin
  NewProgramIndex := 0;
  NewTargetPos := 0;
  SetLength(NewProgram, Length(FProgram));

  FProgramIndex := 0;
  while (FProgramIndex < Length(FProgram)) do begin
    FirstIndex := FProgramIndex;
    case FProgram[FProgramIndex].Command of
      cmPointerInc: begin
        NewProgram[NewProgramIndex].Command := cmPointerInc;
        NewProgram[NewProgramIndex].Parameter := CheckOccurenceSumParam(cmPointerInc);
      end;
      cmPointerDec: begin
        NewProgram[NewProgramIndex].Command := cmPointerDec;
        NewProgram[NewProgramIndex].Parameter := CheckOccurenceSumParam(cmPointerDec);
      end;
      cmInc: begin
        NewProgram[NewProgramIndex].Command := cmInc;
        NewProgram[NewProgramIndex].Parameter := CheckOccurenceSumParam(cmInc);
      end;
      cmDec: begin
        NewProgram[NewProgramIndex].Command := cmDec;
        NewProgram[NewProgramIndex].Parameter := CheckOccurenceSumParam(cmDec);
      end;
      else NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
    end;
    DebugSteps.UpdateTargetPos(FirstIndex, FProgramIndex, NewProgramIndex, NewTargetPos);
    Inc(NewTargetPos, Length(GetOperationText(NewProgram[NewProgramIndex])));
    Inc(FProgramIndex);
    Inc(NewProgramIndex);
  end;
  SetLength(NewProgram, NewProgramIndex);

  // Replace old program by new program
  SetLength(FProgram, Length(NewProgram));
  Move(Pointer(NewProgram)^, Pointer(FProgram)^, SizeOf(TMachineOperation) * Length(NewProgram));
end;

procedure TBFTarget.OptimizeMerge;
var
  NewProgram: array of TMachineOperation;
  NewProgramIndex: Integer;
  PreviousCommand: TMachineCommand;
  FirstIndex: Integer;
  NewTargetIndex: Integer;
begin
  // Merge together cmInc, cmDec, cmSet
  // Merge together cmPointerInc, cmPointerDec
  PreviousCommand := cmNoOperation;
  NewProgramIndex := 0;
  SetLength(NewProgram, Length(FProgram));

  FProgramIndex := 0;
  NewTargetIndex := 0;
  while (FProgramIndex < Length(FProgram)) do begin
    FirstIndex := FProgramIndex;
    case FProgram[FProgramIndex].Command of
      cmPointerInc: begin
        if PreviousCommand in [cmPointerInc, cmPointerDec] then begin
          if NewProgram[NewProgramIndex - 1].Command = cmPointerInc then
            NewProgram[NewProgramIndex - 1].Parameter := NewProgram[NewProgramIndex - 1].Parameter +
              FProgram[FProgramIndex].Parameter
          else if NewProgram[NewProgramIndex - 1].Command = cmPointerDec then
            NewProgram[NewProgramIndex - 1].Parameter := NewProgram[NewProgramIndex - 1].Parameter -
              FProgram[FProgramIndex].Parameter;
          // If value negative then change command
          if NewProgram[NewProgramIndex - 1].Parameter < 0 then begin
            NewProgram[NewProgramIndex - 1].Parameter := -NewProgram[NewProgramIndex - 1].Parameter;
            if NewProgram[NewProgramIndex - 1].Command = cmPointerInc then
              NewProgram[NewProgramIndex - 1].Command := cmPointerDec
              else NewProgram[NewProgramIndex - 1].Command := cmPointerInc;
          end;
          if NewProgram[NewProgramIndex - 1].Parameter = 0 then Dec(NewProgramIndex);
          Dec(NewProgramIndex);
        end else begin
          NewProgram[NewProgramIndex].Command := cmPointerInc;
          NewProgram[NewProgramIndex].Parameter := FProgram[FProgramIndex].Parameter;
        end;
      end;
      cmPointerDec: begin
        if PreviousCommand in [cmPointerInc, cmPointerDec] then begin
          if NewProgram[NewProgramIndex - 1].Command = cmPointerDec then
            NewProgram[NewProgramIndex - 1].Parameter := NewProgram[NewProgramIndex - 1].Parameter +
              FProgram[FProgramIndex].Parameter
          else if NewProgram[NewProgramIndex - 1].Command = cmPointerInc then
            NewProgram[NewProgramIndex - 1].Parameter := NewProgram[NewProgramIndex - 1].Parameter -
              FProgram[FProgramIndex].Parameter;
          // If value negative then change command
          if NewProgram[NewProgramIndex - 1].Parameter < 0 then begin
            NewProgram[NewProgramIndex - 1].Parameter := -NewProgram[NewProgramIndex - 1].Parameter;
            if NewProgram[NewProgramIndex - 1].Command = cmPointerInc then
              NewProgram[NewProgramIndex - 1].Command := cmPointerDec
              else NewProgram[NewProgramIndex - 1].Command := cmPointerInc;
          end;
          if NewProgram[NewProgramIndex - 1].Parameter = 0 then Dec(NewProgramIndex);
          Dec(NewProgramIndex);
        end else begin
          NewProgram[NewProgramIndex].Command := cmPointerDec;
          NewProgram[NewProgramIndex].Parameter := FProgram[FProgramIndex].Parameter;
        end;
      end;
      cmInc: begin
        if PreviousCommand in [cmInc, cmDec, cmSet] then begin
          if NewProgram[NewProgramIndex - 1].Command in [cmInc, cmSet] then
            NewProgram[NewProgramIndex - 1].Parameter := NewProgram[NewProgramIndex - 1].Parameter +
              FProgram[FProgramIndex].Parameter
          else if NewProgram[NewProgramIndex - 1].Command = cmDec then
            NewProgram[NewProgramIndex - 1].Parameter := NewProgram[NewProgramIndex - 1].Parameter -
              FProgram[FProgramIndex].Parameter;
          // If value negative then change command
          if (NewProgram[NewProgramIndex - 1].Parameter < 0) and (NewProgram[NewProgramIndex - 1].Command <> cmSet) then begin
            NewProgram[NewProgramIndex - 1].Parameter := -NewProgram[NewProgramIndex - 1].Parameter;
            if NewProgram[NewProgramIndex - 1].Command = cmInc then
              NewProgram[NewProgramIndex - 1].Command := cmDec
              else NewProgram[NewProgramIndex - 1].Command := cmInc;
          end;
          if NewProgram[NewProgramIndex - 1].Parameter = 0 then Dec(NewProgramIndex);
          Dec(NewProgramIndex);
        end else begin
          NewProgram[NewProgramIndex].Command := cmInc;
          NewProgram[NewProgramIndex].Parameter := FProgram[FProgramIndex].Parameter;
        end;
      end;
      cmDec: begin
        if PreviousCommand in [cmInc, cmDec, cmSet] then begin
          if NewProgram[NewProgramIndex - 1].Command = cmDec then
            NewProgram[NewProgramIndex - 1].Parameter := NewProgram[NewProgramIndex - 1].Parameter +
              FProgram[FProgramIndex].Parameter
          else if NewProgram[NewProgramIndex - 1].Command in [cmInc, cmSet] then
            NewProgram[NewProgramIndex - 1].Parameter := NewProgram[NewProgramIndex - 1].Parameter -
              FProgram[FProgramIndex].Parameter;
          // If value negative then change command
          if (NewProgram[NewProgramIndex - 1].Parameter < 0) and (NewProgram[NewProgramIndex - 1].Command <> cmSet) then begin
            NewProgram[NewProgramIndex - 1].Parameter := -NewProgram[NewProgramIndex - 1].Parameter;
            if NewProgram[NewProgramIndex - 1].Command = cmInc then
              NewProgram[NewProgramIndex - 1].Command := cmDec
              else NewProgram[NewProgramIndex - 1].Command := cmInc;
          end;
          if NewProgram[NewProgramIndex - 1].Parameter = 0 then Dec(NewProgramIndex);
          Dec(NewProgramIndex);
        end else begin
          NewProgram[NewProgramIndex].Command := cmDec;
          NewProgram[NewProgramIndex].Parameter := FProgram[FProgramIndex].Parameter;
        end;
      end;
      cmSet: begin
        if PreviousCommand in [cmInc, cmDec, cmSet] then begin
          // Set overrides value of previous commands
          Dec(NewProgramIndex);
          NewProgram[NewProgramIndex].Command := cmSet;
          NewProgram[NewProgramIndex].Parameter := FProgram[FProgramIndex].Parameter;
        end else begin
          NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
        end;
      end;
      cmLoopStart: begin
        if CheckClear then begin
          NewProgram[NewProgramIndex] := TMachineOperation.Create(cmSet, 0, 0);
          Inc(FProgramIndex, 2);
        end else begin
          NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
        end;
      end;
      else NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
    end;
    PreviousCommand := FProgram[FProgramIndex].Command;
    DebugSteps.UpdateTargetPos(FirstIndex, FProgramIndex, NewProgramIndex, NewTargetIndex);
    Inc(NewTargetIndex, Length(GetOperationText(NewProgram[NewProgramIndex])));
    Inc(FProgramIndex);
    Inc(NewProgramIndex);
  end;
  SetLength(NewProgram, NewProgramIndex);

  // Replace old program by new program
  SetLength(FProgram, Length(NewProgram));
  Move(Pointer(NewProgram)^, Pointer(FProgram)^, SizeOf(TMachineOperation) * Length(NewProgram));
end;

procedure TBFTarget.OptimizeZeroInitMemory;
begin
  // Here Optimizations related to assumption that initial memory is filled with zeroes
  // Then code for constants preparation can be translated to cmSet commands
  // To eliminate also loops for building constants code need to be somehow interpretted partialy
end;

procedure TBFTarget.OptimizeRelativeIndexes;
var
  NewProgram: array of TMachineOperation;
  NewProgramIndex: Integer;
  RelIndex: Integer;
  FirstIndex: Integer;
  NewTargetIndex: Integer;
begin
  NewProgramIndex := 0;
  SetLength(NewProgram, Length(FProgram));

  RelIndex := 0;
  FProgramIndex := 0;
  NewTargetIndex := 0;
  while (FProgramIndex < Length(FProgram)) do begin
    FirstIndex := FProgramIndex;
    case FProgram[FProgramIndex].Command of
      cmPointerInc: begin
        RelIndex := RelIndex + FProgram[FProgramIndex].Parameter;
        Dec(NewProgramIndex);
      end;
      cmPointerDec: begin
        RelIndex := RelIndex - FProgram[FProgramIndex].Parameter;
        Dec(NewProgramIndex);
      end;
      cmInc, cmDec, cmInput, cmOutput, cmSet: begin
        NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
        NewProgram[NewProgramIndex].RelIndex :=
          NewProgram[NewProgramIndex].RelIndex + RelIndex;
      end;
      cmLoopStart, cmLoopEnd: begin
        if RelIndex > 0 then begin
          NewProgram[NewProgramIndex] := TMachineOperation.Create(cmPointerInc,
            RelIndex, 0);
          Inc(NewProgramIndex);
          RelIndex := 0;
        end else
        if RelIndex < 0 then begin
          NewProgram[NewProgramIndex] := TMachineOperation.Create(cmPointerDec,
            Abs(RelIndex), 0);
          Inc(NewProgramIndex);
          RelIndex := 0;
        end;
        NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
      end;
      else raise Exception.Create(Format('Unsupported command %d', [FProgram[FProgramIndex].Command]));
    end;
    DebugSteps.UpdateTargetPos(FirstIndex, FProgramIndex, NewProgramIndex, NewTargetIndex);
    Inc(NewTargetIndex, Length(GetOperationText(NewProgram[NewProgramIndex])));
    Inc(FProgramIndex);
    Inc(NewProgramIndex);
  end;
  SetLength(NewProgram, NewProgramIndex);

  // Replace old program by new program
  SetLength(FProgram, Length(NewProgram));
  Move(Pointer(NewProgram)^, Pointer(FProgram)^, SizeOf(TMachineOperation) *
    Length(NewProgram));
end;

procedure TBFTarget.OptimizeCopyMultiply;
var
  NewProgram: array of TMachineOperation;
  NewProgramIndex: Integer;
  ProcessLoop: Boolean;
  PointerChange: Integer;
  NumberOfBaseDecrement: Integer;
  LoopStartIndex: Integer;
  LoopStartIndexNew: Integer;
  FirstIndex: Integer;
  NewTextIndex: Integer;
begin
  NewProgramIndex := 0;
  SetLength(NewProgram, Length(FProgram));

  NumberOfBaseDecrement := 0;
  ProcessLoop := False;
  FProgramIndex := 0;
  NewTextIndex := 0;
  PointerChange := 0;
  while (FProgramIndex < Length(FProgram)) do begin
    FirstIndex := FProgramIndex;
    case FProgram[FProgramIndex].Command of
      cmPointerInc: begin
        PointerChange := PointerChange + FProgram[FProgramIndex].Parameter;
        NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
      end;
      cmPointerDec: begin
        PointerChange := PointerChange - FProgram[FProgramIndex].Parameter;
        NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
      end;
      cmInc: begin
        if not ProcessLoop then begin
          NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
        end else begin
          if ((FProgram[FProgramIndex].RelIndex + PointerChange) <> 0) then begin
            NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
            NewProgram[NewProgramIndex].Command := cmMultipy;
          end else Dec(NewProgramIndex);
        end;
      end;
      cmDec: begin
        if not ProcessLoop then begin
          if (PointerChange = 0) and (FProgram[FProgramIndex].RelIndex = 0) and
            (FProgram[FProgramIndex].Parameter = 1) then
            Inc(NumberOfBaseDecrement);
          NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
        end else begin
          if ((FProgram[FProgramIndex].RelIndex + PointerChange) <> 0) then begin
            NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
            NewProgram[NewProgramIndex].Command := cmMultipy;
            NewProgram[NewProgramIndex].Parameter := -FProgram[FProgramIndex].Parameter;
          end else Dec(NewProgramIndex);
        end;
      end;
      cmInput, cmOutput: begin
        NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
        Inc(NumberOfBaseDecrement, 2);
      end;
      cmSet: begin
        NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
        Inc(NumberOfBaseDecrement, 2);
      end;
      cmLoopStart: begin
        if not ProcessLoop then begin
          NumberOfBaseDecrement := 0;
          PointerChange := 0;
          LoopStartIndex := FProgramIndex;
          LoopStartIndexNew := NewProgramIndex;
          NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
        end else begin
          Dec(NewProgramIndex);
        end;
      end;
      cmLoopEnd: begin
        if not ProcessLoop then begin
          if (NumberOfBaseDecrement = 1) and (PointerChange = 0) then begin
            FProgramIndex := LoopstartIndex - 1;
            NewProgramIndex := LoopStartIndexNew - 1;
            ProcessLoop := True;
          end else begin
            NewProgram[NewProgramIndex] := FProgram[FProgramIndex];
          end;
        end else begin
          NewProgram[NewProgramIndex] := TMachineOperation.Create(cmSet, 0, 0);
          ProcessLoop := False;
          NumberOfBaseDecrement := 0;
        end;
      end;
      else raise Exception.Create(Format('Unsupported command %d', [FProgram[FProgramIndex].Command]));
    end;
    DebugSteps.UpdateTargetPos(FirstIndex, FProgramIndex, NewProgramIndex, NewTextIndex);
    Inc(NewTextIndex, Length(GetOperationText(NewProgram[NewProgramIndex])));
    Inc(FProgramIndex);
    Inc(NewProgramIndex);
  end;
  SetLength(NewProgram, NewProgramIndex);

  // Replace old program by new program
  SetLength(FProgram, Length(NewProgram));
  Move(Pointer(NewProgram)^, Pointer(FProgram)^, SizeOf(TMachineOperation) *
    Length(NewProgram));
end;

function TBFTarget.GetOperationText(Operation: TMachineOperation): string;
begin
  Result := BrainFuckCommandText[Operation.Command];
  if Operation.Command in [cmInc, cmDec, cmPointerInc, cmPointerDec,
    cmSet, cmMultipy] then begin
  if Operation.Parameter <> 1 then
    Result := Result + IntToStr(Operation.Parameter);
  end;
  if Operation.RelIndex <> 0 then
    Result := Result + 'R' + IntToStr(Operation.RelIndex);
end;

procedure TBFTarget.LoadProgram;
var
  I: Integer;
begin
  inherited;
  DebugSteps.Clear;
  SetLength(FProgram, Length(FSourceCode));
  FProgramIndex := 0;
  for I := 1 to Length(FSourceCode) do begin
    case FSourceCode[I] of
      '+': begin
        FProgram[FProgramIndex] := TMachineOperation.Create(cmInc, 1, 0);
        DebugSteps.AddStep(I - 1, FProgramIndex, soNormal);
      end;
      '-': begin
        FProgram[FProgramIndex] := TMachineOperation.Create(cmDec, 1, 0);
        DebugSteps.AddStep(I - 1, FProgramIndex, soNormal);
      end;
      '>': begin
        FProgram[FProgramIndex] := TMachineOperation.Create(cmPointerInc, 1, 0);
        DebugSteps.AddStep(I - 1, FProgramIndex, soNormal);
      end;
      '<': begin
        FProgram[FProgramIndex] := TMachineOperation.Create(cmPointerDec, 1, 0);
        DebugSteps.AddStep(I - 1, FProgramIndex, soNormal);
      end;
      ',': begin
        FProgram[FProgramIndex] := TMachineOperation.Create(cmInput, 0, 0);
        DebugSteps.AddStep(I - 1, FProgramIndex, soNormal);
      end;
      '.': begin
        FProgram[FProgramIndex] := TMachineOperation.Create(cmOutput, 0, 0);
        DebugSteps.AddStep(I - 1, FProgramIndex, soNormal);
      end;
      '[': begin
        FProgram[FProgramIndex] := TMachineOperation.Create(cmLoopStart, 0, 0);
        DebugSteps.AddStep(I - 1, FProgramIndex, soStepIn);
      end;
      ']': begin
        FProgram[FProgramIndex] := TMachineOperation.Create(cmLoopEnd, 0 ,0);
        DebugSteps.AddStep(I - 1, FProgramIndex, soStepOut);
      end
      else Dec(FProgramIndex);
    end;
    Inc(FProgramIndex);
  end;
  SetLength(FProgram, FProgramIndex);
end;

constructor TBFTarget.Create;
begin
  inherited Create;
  MemorySize := 30000;
  CellSize := 256;
end;

procedure TBFTarget.OptimizeSource;
var
  OldLength: Integer;
begin
  inherited;
  if Optimizations.AddSub then OptimizeAddSub;
  if Optimizations.Merge then
  repeat
    OldLength := Length(FProgram);
    OptimizeMerge;
  until Length(FProgram) = OldLength;
  OptimizeZeroInitMemory;
  if Optimizations.RelativeIndexes then OptimizeRelativeIndexes;
  if Optimizations.CopyMultiply then OptimizeCopyMultiply;
end;



end.

